"""Unittests for analytics_enrollments.py"""

from datetime import datetime

import ddt
from mock import patch
from rest_framework import status
from rest_framework.reverse import reverse
from rest_framework.test import APITransactionTestCase

from enterprise_data.admin_analytics.constants import EnrollmentChart, ResponseType
from enterprise_data.api.v1.serializers import AdvanceAnalyticsEnrollmentStatsSerializer as EnrollmentStatsSerializer
from enterprise_data.api.v1.serializers import AdvanceAnalyticsQueryParamSerializer
from enterprise_data.tests.admin_analytics.mock_analytics_data import (
    ENROLLMENT_STATS_CSVS,
    ENROLLMENTS,
    enrollments_csv_content,
    enrollments_dataframe,
)
from enterprise_data.tests.mixins import JWTTestMixin
from enterprise_data.tests.test_utils import UserFactory
from enterprise_data_roles.constants import ENTERPRISE_DATA_ADMIN_ROLE
from enterprise_data_roles.models import EnterpriseDataFeatureRole, EnterpriseDataRoleAssignment

INVALID_CALCULATION_ERROR = (
    f"Calculation must be one of {AdvanceAnalyticsQueryParamSerializer.CALCULATION_CHOICES}"
)
INVALID_GRANULARITY_ERROR = (
    f"Granularity must be one of {AdvanceAnalyticsQueryParamSerializer.GRANULARITY_CHOICES}"
)
INVALID_RESPONSE_TYPE_ERROR = f"response_type must be one of {AdvanceAnalyticsQueryParamSerializer.RESPONSE_TYPES}"
INVALID_CHART_TYPE_ERROR = f"chart_type must be one of {EnrollmentStatsSerializer.CHART_TYPES}"


@ddt.ddt
class TestIndividualEnrollmentsAPI(JWTTestMixin, APITransactionTestCase):
    """Tests for AdvanceAnalyticsIndividualEnrollmentsView."""

    def setUp(self):
        """
        Setup method.
        """
        super().setUp()
        self.user = UserFactory(is_staff=True)
        role, __ = EnterpriseDataFeatureRole.objects.get_or_create(
            name=ENTERPRISE_DATA_ADMIN_ROLE
        )
        self.role_assignment = EnterpriseDataRoleAssignment.objects.create(
            role=role, user=self.user
        )
        self.client.force_authenticate(user=self.user)

        self.enterprise_uuid = "ee5e6b3a-069a-4947-bb8d-d2dbc323396c"
        self.set_jwt_cookie()

        self.url = reverse(
            "v1:enterprise-admin-analytics-enrollments",
            kwargs={"enterprise_uuid": self.enterprise_uuid},
        )

        fetch_max_enrollment_datetime_patcher = patch(
            'enterprise_data.api.v1.views.analytics_enrollments.fetch_enrollments_cache_expiry_timestamp',
            return_value=datetime.now()
        )

        fetch_max_enrollment_datetime_patcher.start()
        self.addCleanup(fetch_max_enrollment_datetime_patcher.stop)

    def verify_enrollment_data(self, results, results_count):
        """Verify the received enrollment data."""
        attrs = [
            "email",
            "course_title",
            "course_subject",
            "enroll_type",
            "enterprise_enrollment_date",
        ]

        assert len(results) == results_count

        filtered_data = []
        for enrollment in ENROLLMENTS:
            for result in results:
                if enrollment["email"] == result["email"]:
                    filtered_data.append({attr: enrollment[attr] for attr in attrs})
                    break

        received_data = sorted(results, key=lambda x: x["email"])
        expected_data = sorted(filtered_data, key=lambda x: x["email"])
        assert received_data == expected_data

    @patch(
        "enterprise_data.api.v1.views.analytics_enrollments.fetch_and_cache_enrollments_data"
    )
    def test_get(self, mock_fetch_and_cache_enrollments_data):
        """
        Test the GET method for the AdvanceAnalyticsIndividualEnrollmentsView works.
        """
        mock_fetch_and_cache_enrollments_data.return_value = enrollments_dataframe()

        response = self.client.get(self.url, {"page_size": 2})
        assert response.status_code == status.HTTP_200_OK
        assert response["Content-Type"] == "application/json"
        data = response.json()
        assert data["next"] == f"http://testserver{self.url}?page=2&page_size=2"
        assert data["previous"] is None
        assert data["current_page"] == 1
        assert data["num_pages"] == 3
        assert data["count"] == 5
        self.verify_enrollment_data(data["results"], 2)

        response = self.client.get(self.url, {"page_size": 2, "page": 2})
        assert response.status_code == status.HTTP_200_OK
        assert response["Content-Type"] == "application/json"
        data = response.json()
        assert data["next"] == f"http://testserver{self.url}?page=3&page_size=2"
        assert data["previous"] == f"http://testserver{self.url}?page_size=2"
        assert data["current_page"] == 2
        assert data["num_pages"] == 3
        assert data["count"] == 5
        self.verify_enrollment_data(data["results"], 2)

        response = self.client.get(self.url, {"page_size": 2, "page": 3})
        assert response.status_code == status.HTTP_200_OK
        assert response["Content-Type"] == "application/json"
        data = response.json()
        assert data["next"] is None
        assert data["previous"] == f"http://testserver{self.url}?page=2&page_size=2"
        assert data["current_page"] == 3
        assert data["num_pages"] == 3
        assert data["count"] == 5
        self.verify_enrollment_data(data["results"], 1)

        response = self.client.get(self.url, {"page_size": 5})
        assert response.status_code == status.HTTP_200_OK
        assert response["Content-Type"] == "application/json"
        data = response.json()
        assert data["next"] is None
        assert data["previous"] is None
        assert data["current_page"] == 1
        assert data["num_pages"] == 1
        assert data["count"] == 5
        self.verify_enrollment_data(data["results"], 5)

    @patch(
        "enterprise_data.api.v1.views.analytics_enrollments.fetch_and_cache_enrollments_data"
    )
    def test_get_csv(self, mock_fetch_and_cache_enrollments_data):
        """
        Test the GET method for the AdvanceAnalyticsIndividualEnrollmentsView return correct CSV data.
        """
        mock_fetch_and_cache_enrollments_data.return_value = enrollments_dataframe()
        response = self.client.get(self.url, {"response_type": ResponseType.CSV.value})
        assert response.status_code == status.HTTP_200_OK

        # verify the response headers
        assert response["Content-Type"] == "text/csv"

        # verify the response content
        content = b"".join(response.streaming_content)
        assert content == enrollments_csv_content()

    @ddt.data(
        {
            "params": {"start_date": 1},
            "error": {
                "start_date": [
                    "Date has wrong format. Use one of these formats instead: YYYY-MM-DD."
                ]
            },
        },
        {
            "params": {"end_date": 2},
            "error": {
                "end_date": [
                    "Date has wrong format. Use one of these formats instead: YYYY-MM-DD."
                ]
            },
        },
        {
            "params": {"start_date": "2024-01-01", "end_date": "2023-01-01"},
            "error": {
                "non_field_errors": [
                    "start_date should be less than or equal to end_date."
                ]
            },
        },
        {
            "params": {"calculation": "invalid"},
            "error": {"calculation": [INVALID_CALCULATION_ERROR]},
        },
        {
            "params": {"granularity": "invalid"},
            "error": {"granularity": [INVALID_GRANULARITY_ERROR]},
        },
        {"params": {"response_type": "invalid"}, "error": {"response_type": [INVALID_RESPONSE_TYPE_ERROR]}},
    )
    @ddt.unpack
    def test_get_invalid_query_params(self, params, error):
        """
        Test the GET method return correct error if any query param value is incorrect.
        """
        response = self.client.get(self.url, params)
        assert response.status_code == status.HTTP_400_BAD_REQUEST
        assert response.json() == error


@ddt.ddt
class TestEnrollmentStatsAPI(JWTTestMixin, APITransactionTestCase):
    """Tests for AdvanceAnalyticsEnrollmentStatsView."""

    def setUp(self):
        """
        Setup method.
        """
        super().setUp()
        self.user = UserFactory(is_staff=True)
        role, __ = EnterpriseDataFeatureRole.objects.get_or_create(
            name=ENTERPRISE_DATA_ADMIN_ROLE
        )
        self.role_assignment = EnterpriseDataRoleAssignment.objects.create(
            role=role, user=self.user
        )
        self.client.force_authenticate(user=self.user)

        self.enterprise_uuid = "ee5e6b3a-069a-4947-bb8d-d2dbc323396c"
        self.set_jwt_cookie()

        self.url = reverse(
            "v1:enterprise-admin-analytics-enrollments-stats",
            kwargs={"enterprise_uuid": self.enterprise_uuid},
        )

        fetch_max_enrollment_datetime_patcher = patch(
            'enterprise_data.api.v1.views.analytics_enrollments.fetch_enrollments_cache_expiry_timestamp',
            return_value=datetime.now()
        )

        fetch_max_enrollment_datetime_patcher.start()
        self.addCleanup(fetch_max_enrollment_datetime_patcher.stop)

    def verify_enrollment_data(self, results):
        """Verify the received enrollment data."""
        attrs = [
            "email",
            "course_title",
            "course_subject",
            "enroll_type",
            "enterprise_enrollment_date",
        ]

        filtered_data = []
        for enrollment in ENROLLMENTS:
            for result in results:
                if enrollment["email"] == result["email"]:
                    filtered_data.append({attr: enrollment[attr] for attr in attrs})
                    break

        received_data = sorted(results, key=lambda x: x["email"])
        expected_data = sorted(filtered_data, key=lambda x: x["email"])
        assert received_data == expected_data

    @patch(
        "enterprise_data.api.v1.views.analytics_enrollments.fetch_and_cache_enrollments_data"
    )
    def test_get(self, mock_fetch_and_cache_enrollments_data):
        """
        Test the GET method for the AdvanceAnalyticsEnrollmentStatsView works.
        """
        mock_fetch_and_cache_enrollments_data.return_value = enrollments_dataframe()

        response = self.client.get(self.url)
        assert response.status_code == status.HTTP_200_OK
        assert response["Content-Type"] == "application/json"
        data = response.json()
        assert data == {
            "enrollments_over_time": [
                {
                    "enterprise_enrollment_date": "2020-04-03T00:00:00",
                    "enroll_type": "certificate",
                    "count": 1,
                },
                {
                    "enterprise_enrollment_date": "2020-04-08T00:00:00",
                    "enroll_type": "certificate",
                    "count": 1,
                },
                {
                    "enterprise_enrollment_date": "2021-05-11T00:00:00",
                    "enroll_type": "certificate",
                    "count": 1,
                },
                {
                    "enterprise_enrollment_date": "2021-07-03T00:00:00",
                    "enroll_type": "certificate",
                    "count": 1,
                },
                {
                    "enterprise_enrollment_date": "2021-07-04T00:00:00",
                    "enroll_type": "certificate",
                    "count": 1,
                },
            ],
            "top_courses_by_enrollments": [
                {"course_key": "NOGk+UVD31", "enroll_type": "certificate", "count": 1},
                {"course_key": "QWXx+Jqz64", "enroll_type": "certificate", "count": 1},
                {"course_key": "hEmW+tvk03", "enroll_type": "certificate", "count": 2},
                {"course_key": "qZJC+KFX86", "enroll_type": "certificate", "count": 1},
            ],
            "top_subjects_by_enrollments": [
                {
                    "course_subject": "business-management",
                    "enroll_type": "certificate",
                    "count": 2,
                },
                {
                    "course_subject": "communication",
                    "enroll_type": "certificate",
                    "count": 1,
                },
                {
                    "course_subject": "medicine",
                    "enroll_type": "certificate",
                    "count": 1,
                },
                {
                    "course_subject": "social-sciences",
                    "enroll_type": "certificate",
                    "count": 1,
                },
            ],
        }

    @patch("enterprise_data.api.v1.views.analytics_enrollments.fetch_and_cache_enrollments_data")
    @ddt.data(
        EnrollmentChart.ENROLLMENTS_OVER_TIME.value,
        EnrollmentChart.TOP_COURSES_BY_ENROLLMENTS.value,
        EnrollmentChart.TOP_SUBJECTS_BY_ENROLLMENTS.value,
    )
    def test_get_csv(self, chart_type, mock_fetch_and_cache_enrollments_data):
        """
        Test that AdvanceAnalyticsEnrollmentStatsView return correct CSV data.
        """
        mock_fetch_and_cache_enrollments_data.return_value = enrollments_dataframe()

        response = self.client.get(self.url, {"response_type": ResponseType.CSV.value, "chart_type": chart_type})
        assert response.status_code == status.HTTP_200_OK
        assert response["Content-Type"] == "text/csv"
        # verify the response content
        assert response.content == ENROLLMENT_STATS_CSVS[chart_type]

    @ddt.data(
        {
            "params": {"start_date": 1},
            "error": {
                "start_date": [
                    "Date has wrong format. Use one of these formats instead: YYYY-MM-DD."
                ]
            },
        },
        {
            "params": {"end_date": 2},
            "error": {
                "end_date": [
                    "Date has wrong format. Use one of these formats instead: YYYY-MM-DD."
                ]
            },
        },
        {
            "params": {"start_date": "2024-01-01", "end_date": "2023-01-01"},
            "error": {
                "non_field_errors": [
                    "start_date should be less than or equal to end_date."
                ]
            },
        },
        {
            "params": {"calculation": "invalid"},
            "error": {"calculation": [INVALID_CALCULATION_ERROR]},
        },
        {
            "params": {"granularity": "invalid"},
            "error": {"granularity": [INVALID_GRANULARITY_ERROR]},
        },
        {"params": {"chart_type": "invalid"}, "error": {"chart_type": [INVALID_CHART_TYPE_ERROR]}},
    )
    @ddt.unpack
    def test_get_invalid_query_params(self, params, error):
        """
        Test the GET method return correct error if any query param value is incorrect.
        """
        response = self.client.get(self.url, params)
        assert response.status_code == status.HTTP_400_BAD_REQUEST
        assert response.json() == error



"""Advance Analytics for Enrollments"""
from datetime import datetime

from edx_rbac.decorators import permission_required
from edx_rest_framework_extensions.auth.jwt.authentication import JwtAuthentication
from rest_framework.response import Response
from rest_framework.views import APIView

from django.http import HttpResponse, StreamingHttpResponse

from enterprise_data.admin_analytics.constants import Calculation, EnrollmentChart, Granularity, ResponseType
from enterprise_data.admin_analytics.utils import (
    calculation_aggregation,
    fetch_and_cache_enrollments_data,
    fetch_enrollments_cache_expiry_timestamp,
    granularity_aggregation,
)
from enterprise_data.api.v1.paginators import AdvanceAnalyticsPagination
from enterprise_data.api.v1.serializers import (
    AdvanceAnalyticsEnrollmentStatsSerializer,
    AdvanceAnalyticsQueryParamSerializer,
)
from enterprise_data.renderers import IndividualEnrollmentsCSVRenderer
from enterprise_data.utils import date_filter


class AdvanceAnalyticsIndividualEnrollmentsView(APIView):
    """
    API for getting the advance analytics individual enrollments data.
    """
    authentication_classes = (JwtAuthentication,)
    pagination_class = AdvanceAnalyticsPagination
    http_method_names = ['get']

    @permission_required('can_access_enterprise', fn=lambda request, enterprise_uuid: enterprise_uuid)
    def get(self, request, enterprise_uuid):
        """Get individual enrollments data"""
        serializer = AdvanceAnalyticsQueryParamSerializer(data=request.GET)
        serializer.is_valid(raise_exception=True)

        cache_expiry = fetch_enrollments_cache_expiry_timestamp()
        enrollments_df = fetch_and_cache_enrollments_data(enterprise_uuid, cache_expiry)

        # get values from query params or use default values
        start_date = serializer.data.get('start_date', enrollments_df.enterprise_enrollment_date.min())
        end_date = serializer.data.get('end_date', datetime.now())
        response_type = request.query_params.get('response_type', ResponseType.JSON.value)

        # filter enrollments by date
        enrollments = date_filter(start_date, end_date, enrollments_df, "enterprise_enrollment_date")

        # select only the columns that will be in the table.
        enrollments = enrollments[
            [
                "email",
                "course_title",
                "course_subject",
                "enroll_type",
                "enterprise_enrollment_date",
            ]
        ]
        enrollments["enterprise_enrollment_date"] = enrollments["enterprise_enrollment_date"].dt.date
        enrollments = enrollments.sort_values(by="enterprise_enrollment_date", ascending=False).reset_index(drop=True)

        if response_type == ResponseType.CSV.value:
            filename = f"""individual_enrollments, {start_date} - {end_date}.csv"""
            return StreamingHttpResponse(
                IndividualEnrollmentsCSVRenderer().render(self._stream_serialized_data(enrollments)),
                content_type="text/csv",
                headers={"Content-Disposition": f'attachment; filename="{filename}"'},
            )

        paginator = self.pagination_class()
        page = paginator.paginate_queryset(enrollments, request)
        serialized_data = page.data.to_dict(orient='records')
        response = paginator.get_paginated_response(serialized_data)

        return response

    def _stream_serialized_data(self, enrollments, chunk_size=50000):
        """
        Stream the serialized data.
        """
        total_rows = enrollments.shape[0]
        for start_index in range(0, total_rows, chunk_size):
            end_index = min(start_index + chunk_size, total_rows)
            chunk = enrollments.iloc[start_index:end_index]
            yield from chunk.to_dict(orient='records')


class AdvanceAnalyticsEnrollmentStatsView(APIView):
    """
    API for getting the advance analytics enrollment chart stats.
    """
    authentication_classes = (JwtAuthentication,)
    http_method_names = ['get']

    @permission_required('can_access_enterprise', fn=lambda request, enterprise_uuid: enterprise_uuid)
    def get(self, request, enterprise_uuid):  # lint-amnesty, pylint: disable=inconsistent-return-statements
        """Get enrollment chart stats"""
        serializer = AdvanceAnalyticsEnrollmentStatsSerializer(data=request.GET)
        serializer.is_valid(raise_exception=True)

        cache_expiry = fetch_enrollments_cache_expiry_timestamp()
        enrollments_df = fetch_and_cache_enrollments_data(enterprise_uuid, cache_expiry)

        # get values from query params or use default
        start_date = serializer.data.get('start_date', enrollments_df.enterprise_enrollment_date.min())
        end_date = serializer.data.get('end_date', datetime.now())
        granularity = serializer.data.get('granularity', Granularity.DAILY.value)
        calculation = serializer.data.get('calculation', Calculation.TOTAL.value)
        response_type = serializer.data.get('response_type', ResponseType.JSON.value)
        chart_type = serializer.data.get('chart_type')

        if response_type == ResponseType.JSON.value:
            data = {
                "enrollments_over_time": self.construct_enrollments_over_time(
                    enrollments_df.copy(),
                    start_date,
                    end_date,
                    granularity,
                    calculation,
                ),
                "top_courses_by_enrollments": self.construct_top_courses_by_enrollments(
                    enrollments_df.copy(),
                    start_date,
                    end_date,
                ),
                "top_subjects_by_enrollments": self.construct_top_subjects_by_enrollments(
                    enrollments_df.copy(),
                    start_date,
                    end_date,
                ),
            }
            return Response(data)

        if response_type == ResponseType.CSV.value:
            if chart_type == EnrollmentChart.ENROLLMENTS_OVER_TIME.value:
                return self.construct_enrollments_over_time_csv(
                    enrollments_df.copy(),
                    start_date,
                    end_date,
                    granularity,
                    calculation,
                )
            elif chart_type == EnrollmentChart.TOP_COURSES_BY_ENROLLMENTS.value:
                return self.construct_top_courses_by_enrollments_csv(
                    enrollments_df.copy(),
                    start_date,
                    end_date,
                )
            elif chart_type == EnrollmentChart.TOP_SUBJECTS_BY_ENROLLMENTS.value:
                return self.construct_top_subjects_by_enrollments_csv(
                    enrollments_df.copy(),
                    start_date,
                    end_date,
                )

    def enrollments_over_time_common(self, enrollments_df, start_date, end_date, granularity, calculation):
        """
        Common method for constructing enrollments over time data.

        Arguments:
            enrollments_df {DataFrame} -- DataFrame of enrollments
            start_date {datetime} -- Enrollment start date in the format 'YYYY-MM-DD'
            end_date {datetime} -- Enrollment end date in the format 'YYYY-MM-DD'
            granularity {str} -- Granularity of the data. One of Granularity choices
            calculation {str} -- Calculation of the data. One of Calculation choices
        """
        # filter enrollments by date
        enrollments = date_filter(start_date, end_date, enrollments_df, "enterprise_enrollment_date")

        # aggregate enrollments by granularity
        enrollments = granularity_aggregation(
            level=granularity,
            group=["enterprise_enrollment_date", "enroll_type"],
            date="enterprise_enrollment_date",
            data_frame=enrollments,
        )

        # aggregate enrollments by calculation
        enrollments = calculation_aggregation(calc=calculation, data_frame=enrollments)

        return enrollments

    def construct_enrollments_over_time(self, enrollments_df, start_date, end_date, granularity, calculation):
        """
        Construct enrollments over time data.

        Arguments:
            enrollments_df {DataFrame} -- DataFrame of enrollments
            start_date {datetime} -- Enrollment start date in the format 'YYYY-MM-DD'
            end_date {datetime} -- Enrollment end date in the format 'YYYY-MM-DD'
            granularity {str} -- Granularity of the data. One of Granularity choices
            calculation {str} -- Calculation of the data. One of Calculation choices
        """
        enrollments = self.enrollments_over_time_common(enrollments_df, start_date, end_date, granularity, calculation)

        # convert dataframe to a list of records
        return enrollments.to_dict(orient='records')

    def construct_enrollments_over_time_csv(self, enrollments_df, start_date, end_date, granularity, calculation):
        """
        Construct enrollments over time CSV.

        Arguments:
            enrollments_df {DataFrame} -- DataFrame of enrollments
            start_date {datetime} -- Enrollment start date in the format 'YYYY-MM-DD'
            end_date {datetime} -- Enrollment end date in the format 'YYYY-MM-DD'
            granularity {str} -- Granularity of the data. One of Granularity choices
            calculation {str} -- Calculation of the data. One of Calculation choices
        """
        enrollments = self.enrollments_over_time_common(enrollments_df, start_date, end_date, granularity, calculation)

        enrollments = enrollments.pivot(
            index="enterprise_enrollment_date", columns="enroll_type", values="count"
        )

        filename = f"Enrollment Timeseries, {start_date} - {end_date} ({granularity} {calculation}).csv"
        return self.construct_csv_response(enrollments, filename)

    def top_courses_by_enrollments_common(self, enrollments_df, start_date, end_date, group_by_columns, columns):
        """
        Common method for constructing top courses by enrollments data.

        Arguments:
            enrollments_df {DataFrame} -- DataFrame of enrollments
            start_date {datetime} -- Enrollment start date in the format 'YYYY-MM-DD'
            end_date {datetime} -- Enrollment end date in the format 'YYYY-MM-DD'
            group_by_columns {list} -- List of columns to group by
            columns {list} -- List of column for the final result
        """
        # filter enrollments by date
        enrollments = date_filter(start_date, end_date, enrollments_df, "enterprise_enrollment_date")

        courses = list(
            enrollments.groupby(["course_key"]).size().sort_values(ascending=False)[:10].index
        )

        enrollments = (
            enrollments[enrollments.course_key.isin(courses)]
            .groupby(group_by_columns)
            .size()
            .reset_index()
        )
        enrollments.columns = columns

        return enrollments

    def construct_top_courses_by_enrollments(self, enrollments_df, start_date, end_date):
        """
        Construct top courses by enrollments data.

        Arguments:
            enrollments_df {DataFrame} -- DataFrame of enrollments
            start_date {datetime} -- Enrollment start date in the format 'YYYY-MM-DD'
            end_date {datetime} -- Enrollment end date in the format 'YYYY-MM-DD'
        """
        group_by_columns = ["course_key", "enroll_type"]
        columns = ["course_key", "enroll_type", "count"]
        enrollments = self.top_courses_by_enrollments_common(
            enrollments_df,
            start_date,
            end_date,
            group_by_columns,
            columns
        )

        # convert dataframe to a list of records
        return enrollments.to_dict(orient='records')

    def construct_top_courses_by_enrollments_csv(self, enrollments_df, start_date, end_date):
        """
        Construct top courses by enrollments CSV.

        Arguments:
            enrollments_df {DataFrame} -- DataFrame of enrollments
            start_date {datetime} -- Enrollment start date in the format 'YYYY-MM-DD'
            end_date {datetime} -- Enrollment end date in the format 'YYYY-MM-DD'
        """
        group_by_columns = ["course_key", "course_title", "enroll_type"]
        columns = ["course_key", "course_title", "enroll_type", "count"]
        enrollments = self.top_courses_by_enrollments_common(
            enrollments_df,
            start_date,
            end_date,
            group_by_columns,
            columns
        )

        enrollments = enrollments.pivot(
            index=["course_key", "course_title"], columns="enroll_type", values="count"
        )

        filename = f"Top 10 Courses, {start_date} - {end_date}.csv"
        return self.construct_csv_response(enrollments, filename)

    def top_subjects_by_enrollments_common(self, enrollments_df, start_date, end_date):
        """
        Common method for constructing top subjects by enrollments data.

        Arguments:
            enrollments_df {DataFrame} -- DataFrame of enrollments
            start_date {datetime} -- Enrollment start date in the format 'YYYY-MM-DD'
            end_date {datetime} -- Enrollment end date in the format 'YYYY-MM-DD'
        """
        # filter enrollments by date
        enrollments = date_filter(start_date, end_date, enrollments_df, "enterprise_enrollment_date")

        subjects = list(
            enrollments.groupby(["course_subject"]).size().sort_values(ascending=False)[:10].index
        )

        enrollments = (
            enrollments[enrollments.course_subject.isin(subjects)]
            .groupby(["course_subject", "enroll_type"])
            .size()
            .reset_index()
        )
        enrollments.columns = ["course_subject", "enroll_type", "count"]

        return enrollments

    def construct_top_subjects_by_enrollments(self, enrollments_df, start_date, end_date):
        """
        Construct top subjects by enrollments data.

        Arguments:
            enrollments_df {DataFrame} -- DataFrame of enrollments
            start_date {datetime} -- Enrollment start date in the format 'YYYY-MM-DD'
            end_date {datetime} -- Enrollment end date in the format 'YYYY-MM-DD'
        """
        enrollments = self.top_subjects_by_enrollments_common(enrollments_df, start_date, end_date)
        # convert dataframe to a list of records
        return enrollments.to_dict(orient='records')

    def construct_top_subjects_by_enrollments_csv(self, enrollments_df, start_date, end_date):
        """
        Construct top subjects by enrollments CSV.

        Arguments:
            enrollments_df {DataFrame} -- DataFrame of enrollments
            start_date {datetime} -- Enrollment start date in the format 'YYYY-MM-DD'
            end_date {datetime} -- Enrollment end date in the format 'YYYY-MM-DD'
        """
        enrollments = self.top_subjects_by_enrollments_common(enrollments_df, start_date, end_date)
        enrollments = enrollments.pivot(index="course_subject", columns="enroll_type", values="count")

        filename = f"Top 10 Subjects by Enrollment, {start_date} - {end_date}.csv"
        return self.construct_csv_response(enrollments, filename)

    def construct_csv_response(self, enrollments, filename):
        """
        Construct CSV response.

        Arguments:
            enrollments {DataFrame} -- DataFrame of enrollments
            filename {str} -- Filename for the CSV
        """
        response = HttpResponse(content_type='text/csv')
        response['Content-Disposition'] = f'attachment; filename="{filename}"'
        enrollments.to_csv(path_or_buf=response)

        return response
